import logging
import os
import json
from collections.abc import Mapping
import numpy as np
import pandas as pd
from scipy import integrate
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector, Index, Status
from timewise_sup.meta_analysis.luminosity import get_ir_luminosities


logger = logging.getLogger(__name__)


def ir_energy_integral(
        lc: pd.DataFrame,
        t0: Mapping[str, float],
        t1: Mapping[str, float],

) -> float:

    mjd = lc["mean_mjd"]
    bol_eng = 0

    for b in ["W1", "W2"]:
        mask = (mjd >= t0[b]) & (mjd <= t1[b])
        fit_lum = lc[f"{b}_ir_luminosity_erg_per_s"][mask]
        fit_mjd = mjd[mask]

        # integrate using trapezoidal rule
        integral = integrate.trapezoid(fit_lum, fit_mjd * 24 * 3600)
        bol_eng += integral

    return bol_eng


def calculate_ir_energy(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
) -> dict:
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(status)
    indices = np.atleast_1d(ids)
    logger.info(f"calculating luminosities for {len(indices)} objects")
    lcs = get_ir_luminosities(base_name, database_name, wise_data, status)
    excess_mjd = database_connector.get_excess_mjd(indices)

    engs = dict()

    for k, lc in lcs.items():
        ieng = dict()
        ieng["ir_energy_erg"] = ir_energy_integral(pd.DataFrame.from_dict(lc, orient="columns"),
                                                      {b: excess_mjd[k][f"{b}_excess_start_mjd"] for b in ["W1", "W2"]},
                                                      {b: excess_mjd[k][f"{b}_excess_end_mjd"] for b in ["W1", "W2"]})
        ieng["ir_energy_is_lower_limit"] = bool(np.any([excess_mjd[k][f"{b}_flare_ended"] for b in ["W1", "W2"]]))
        engs[k] = ieng

    return engs


def get_ir_energy_status(
        base_name: str,
        database_name: str,
        status: Status,
        wise_data: WISEDataBase | None = None,
        force_new: bool = False,
) -> dict:
    logger.info(f"getting bolometric luminosities for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"ir_eng_status{status}.json")

    if (not os.path.isfile(fn)) or force_new:

        if wise_data is None:
            raise ValueError("wise_data must be given when calculating IR energies!")
        else:
            logger.debug(f"No file {fn}.")
            logger.info("Making bolometric luminosities")
            engs = calculate_ir_energy(
                base_name=base_name,
                database_name=database_name,
                wise_data=wise_data,
                status=status
            )

        with open(fn, "w") as f:
            json.dump(engs, f)

    else:
        logger.debug(f"loading {fn}")
        with open(fn, "r") as f:
            engs = json.load(f)

    return engs


def get_ir_energy_index(
        base_name: str,
        database_name: str,
        index: Index,
        forcenew: bool = False,
) -> dict:

    statuses = DatabaseConnector(base_name=base_name, database_name=database_name).get_status(index)

    engs_all = dict()

    for status in statuses.status.unique():
        eng = get_ir_energy_status(base_name, database_name, status, forcenew)
        selected_lums = {k: v for k, v in eng.items() if k in statuses.index[statuses.status == status]}
        engs_all.update(selected_lums)

    return engs_all



import logging
import inspect
from pydantic import validator
from typing import Callable, Any
from timewise.config_loader import TimewiseConfig, TimewiseConfigLoader

from timewise_sup.ampel_conf import create_ampel_config_file
from timewise_sup.analyse_lightcurves.bayesian_blocks import bayesian_blocks
from timewise_sup.meta_analysis.baseline_subtraction import get_baseline_subtracted_lightcurves
from timewise_sup.plots.plot_lightcurves import plot_sample_lightcurves
from timewise_sup.meta_analysis.ztf import download_ztffp_per_status


logger = logging.getLogger(__name__)


functions = {
    "create_ampel_config_file": create_ampel_config_file,
    "bayesian_blocks": bayesian_blocks,
    "baseline_subtraction": get_baseline_subtracted_lightcurves,
    "plot": plot_sample_lightcurves,
    "download_ztffp_per_status": download_ztffp_per_status
}


class TimewiseSUPConfigLoader(TimewiseConfigLoader):

    timewise_sup_instructions: list[dict] = list()
    database_name: str

    def parse_config(self):
        timewise_config = super().parse_config()
        return TimewiseSUPConfig(
            database_name=self.database_name,
            timewise_sup_instructions=self.timewise_sup_instructions,
            **timewise_config.dict()
        )


class TimewiseSUPConfig(TimewiseConfig):

    timewise_sup_instructions: list[dict] = list()
    database_name: str

    @validator("timewise_sup_instructions")
    def validate_timewise_sup_instructions(cls, v: list[dict]):
        for instructions in v:
            for fct_name, arguments in instructions.items():

                if fct_name not in functions:
                    available = ", ".join(list(functions.keys()))
                    raise ValueError(f"timewise-sup has no function {fct_name}! Must be either of {available}")

                fct = functions[fct_name]
                # this is a mypy bug: https://github.com/python/mypy/issues/10740

                signature = inspect.signature(fct)  # type: ignore
                param_list = list(signature.parameters)
                # check if the function needs base_name
                for k in ["base_name", "database_name", "wise_data"]:
                    if k in param_list:
                        # enter dummy string in arguments
                        arguments[k] = ""
                # check validity of arguments
                try:
                    _arguments = arguments or dict()  # supply empty dict if arguments is None
                    signature.bind(**_arguments)
                except TypeError as e:
                    raise ValueError(f"{fct_name}: {e}!")

        return v

    def run_config(self):
        logger.info("running config")
        super().run_config()

        for instructions in self.timewise_sup_instructions:
            for fct_name, arguments in instructions.items():
                _arguments = arguments or dict()
                fct = functions[fct_name]
                params = list(inspect.signature(fct).parameters)

                if "base_name" in params:
                    arguments["base_name"] = self.wise_data.base_name
                if "database_name" in params:
                    arguments["database_name"] = self.database_name
                if "wise_data" in params:
                    arguments["wise_data"] = self.wise_data

                _arguments = arguments or dict()  # supply empty dict if arguments is None
                logger.info(f"running {fct_name} with arguments {_arguments}")
                fct(**_arguments)

        logger.info("successfully ran config")



import os.path

import matplotlib.pyplot as plt
import matplotlib as mpl
import logging
import numpy as np
import pandas as pd
from timewise.wise_data_base import WISEDataBase

from timewise_sup.meta_analysis.luminosity import nuFnu
from timewise_sup.meta_analysis.ztf import get_ztf_lightcurve, ztf_start
from timewise_sup.meta_analysis.baseline_subtraction import get_single_lightcurve, get_baseline_subtracted_lightcurves
from timewise_sup.plots import plots_dir, bandcolors


logger = logging.getLogger(__name__)


def plot_sample_lightcurves(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: str,
        include_ztffph: bool = False,
        service: str = "tap",
        load_from_bigdata_dir: bool = False
):
    wise_lcs = get_baseline_subtracted_lightcurves(
        base_name=base_name,
        database_name=database_name,
        wise_data=wise_data,
        status=status,
        service=service,
        load_from_bigdata_dir=load_from_bigdata_dir
    )

    for index, lc_dict in wise_lcs.items():
        logger.debug(f"index {index}")
        wise_lc = pd.DataFrame.from_dict(lc_dict)
        ztf_lc = get_ztf_lightcurve(base_name=base_name, index=index) if include_ztffph else None

        fig, ax = make_lightcurve_plot(wise_lc, ztf_lc)

        if include_ztffph and (ztf_lc is None):
            ax.annotate("no ZTF data", (0.5, 1), xycoords="subfigure fraction")

        d = plots_dir("baseline_subtracted_lightcurves", base_name)
        fn = os.path.join(d, status, f"{index}.pdf")

        d = os.path.dirname(fn)
        if not os.path.isdir(d):
            os.makedirs(d)

        logger.debug(f"saving under {fn}")
        fig.savefig(fn)
        plt.close()


def plot_single_lightcurve(
        base_name: str,
        index: str,
        include_ztffph: bool = False,
        ztf_snr_threshold: float = 5
) -> tuple[mpl.figure.Figure, mpl.axes.Axes]:

    wise_lc = get_single_lightcurve(base_name=base_name, index=index)
    if include_ztffph:
        ztf_lc = get_ztf_lightcurve(base_name=base_name, index=index, snr_threshold=ztf_snr_threshold)
    else:
        ztf_lc = None

    return make_lightcurve_plot(wise_lc, ztf_lc)


def make_lightcurve_plot(
        wise_lc: pd.DataFrame,
        ztf_lc: pd.DataFrame = None,
        unit: str = "erg s-1 cm-2"
) -> tuple[mpl.figure.Figure, mpl.axes.Axes]:

    fig, ax = plt.subplots()

    for b in ["W1", "W2"]:
        ax.errorbar(
            wise_lc.mean_mjd,
            nuFnu(wise_lc[f"{b}_diff_mean_flux_density"], flux_unit="mJy", band=b, out_unit=unit),
            yerr=nuFnu(wise_lc[f"{b}_diff_flux_density_rms"], flux_unit="mJy", band=b, out_unit=unit),
            marker='s',
            ms=3,
            color=bandcolors[b],
            label=f"WISE {b}",
            ls="",
            zorder=5,
            capsize=2
        )

    if ztf_lc is not None:

        if len(ztf_lc) == 0:
            ax.annotate("no data from ZTF observations", (0.5, 1), xycoords="subfigure fraction", color='red')

        else:

            for b in ["g", "r", "i"]:
                band_mask = ztf_lc["filter"] == f"ZTF_{b}"
                snr_mask = ztf_lc["above_snr"][band_mask]

                f = nuFnu(ztf_lc[band_mask]["flux[Jy]"], flux_unit="Jy", band=f"ZTF_{b}", out_unit=unit)
                ferr = nuFnu(ztf_lc[band_mask]["flux_err[Jy]"], flux_unit="Jy", band=f"ZTF_{b}", out_unit=unit)

                if np.any(snr_mask):
                    ax.errorbar(
                        ztf_lc.obsmjd[band_mask][snr_mask],
                        f[snr_mask],
                        yerr=ferr[snr_mask],
                        marker='o',
                        ms=2,
                        color=bandcolors[f"ZTF_{b}"],
                        label="ZTF " + b,
                        ls="",
                        zorder=4,
                        elinewidth=1
                    )

                if np.any(~snr_mask):
                    ylim = ax.get_ylim()  # upper limits should not adjust the axis limits
                    ax.scatter(
                        ztf_lc.obsmjd[band_mask][~snr_mask],
                        f[~snr_mask] + ferr[~snr_mask],
                        marker="v",
                        s=2,
                        alpha=0.2,
                        color=bandcolors[f"ZTF_{b}"],
                        zorder=2,
                        label=f"ZTF {b} upper limits" if np.all(~snr_mask) else ""
                    )
                    ax.set_ylim(ylim)

        xlim = ax.get_xlim()
        if xlim[0] < ztf_start.mjd < xlim[1]:
            ax.axvline(ztf_start.mjd, ls=":", color="grey")
            ax.annotate(
                "ZTF start", (ztf_start.mjd, ax.get_ylim()[1]),
                xytext=(-2, -2), textcoords='offset points',
                rotation=90, va='top', ha='right', color="grey",
                annotation_clip=False
            )

    ax.set_ylabel(r"$\nu$ F$_{\nu}$ [erg s$^{-1}$ cm$^{-2}$]")
    ax.grid()
    ax.legend()

    return fig, ax



import logging
import pandas as pd
import io
import os
import requests
from requests.auth import HTTPBasicAuth
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy import units as u
from timewise import ParentSampleBase, WISEDataDESYCluster

from timewise_sup.config_loader import TimewiseSUPConfig


logger = logging.getLogger(__name__)


def get_bts_config():
    wise_data = WISEDataDESYCluster(
        base_name="bts_sample",
        parent_sample_class=BTSParentSample,
        n_chunks=1,
        min_sep_arcsec=6
    )
    config = TimewiseSUPConfig(
        database_name=wise_data.base_name,
        wise_data=wise_data
    )
    return config


class BTSParentSample(ParentSampleBase):

    default_keymap = {
        "ra": "ra_deg",
        "dec": "dec_deg",
        "id": "IAUID"
    }

    base_name = "bts_sample"

    url = (
        "https://sites.astro.caltech.edu/ztf/rcf/explorer.php?f=s&"
        "coverage=any&samprcf=y&sampdeep=y&subsample=cantrans&classstring=&classexclude=&quality=y&ztflink=fritz&"
        "startsavedate=&startpeakdate=&startlastdate=&startra=&startdec=&startz=&startdur=&startrise=&startfade=&"
        "startpeakmag=&startlastmag=&startabsmag=&starthostabs=&starthostcol=&startsavevis=&startlatevis=&"
        "startcurrvis=&startb=&startav=&endsavedate=&endpeakdate=&endlastdate=&endra=&enddec=&endz=&enddur=&endrise=&"
        "endfade=&endpeakmag=&endlastmag=&endabsmag=&endhostabs=&endhostcol=&endsavevis=&endlatevis=&endcurrvis=&"
        "endb=&endav=&sort=savedate&format=csv"
    )

    def __init__(self):
        super(BTSParentSample, self).__init__(base_name=BTSParentSample.base_name)

        if not os.path.isfile(self.local_sample_copy):
            self.df = self.make_sample()
            self.save_local()

        self.df = pd.read_csv(
            self.local_sample_copy,
            index_col=0
        )

    @staticmethod
    def make_sample():
        logger.info("downloading BTS sample")

        auth = HTTPBasicAuth("bts", "rcf")
        res = requests.get(BTSParentSample.url, auth=auth)

        if res.status_code != 200:
            msg = f"Trying to load BTS sample resulted in status code {res.status_code}!"
            raise requests.exceptions.ConnectionError(msg)

        else:
            logger.info("download successful")

        df = pd.read_csv(io.StringIO(res.content.decode()))
        logger.debug(f"{len(df)} objects")

        logger.info("converting coordinates to degrees")
        coord = SkyCoord(df["RA"], df["Dec"], unit=(u.hourangle, u.deg))
        df["ra_deg"] = coord.ra.value
        df["dec_deg"] = coord.dec.value

        logger.info("converting times to mjds")
        df["peakt_mjd"] = Time(df["peakt"] + 2458000, format="jd").mjd

        logger.info("done")
        return df



import argparse
import os
import subprocess
import logging
import time
import numpy as np

from timewise_sup.environment import environment_variable_defaults, load_environment
from timewise_sup.ampel_conf import ampel_conf_filename
from timewise_sup.analyse_lightcurves.create_job_file_yaml import make_ampel_job_file


logger = logging.getLogger(__name__)
tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
db_port = load_environment("TIMEWISE_SUP_MONGODB_PORT")


class Submitter:

    username = os.path.basename(os.environ["HOME"])
    status_cmd = f"condor_q {username}"
    scratch_on_nodes = "$TMPDIR"
    python_script = os.path.join(os.path.abspath(os.path.dirname(__file__)), "bayesian_blocks.py")

    def __init__(
            self,
            base_name,
            node_memory: str | None = None,
            manual_submit: bool = False,
            max_materialized_jobs: int = 10,
            **ampel_options
    ):

        self.base_name = base_name
        self.ampel_options = ampel_options

        self.dir = os.path.join(tsup_data_dir, "cluster", base_name)
        self.script_dir = os.path.join(self.dir, "scripts")
        self.log_dir = os.path.join(self.dir, "logs")
        for d in [self.dir, self.script_dir, self.log_dir]:
            if not os.path.isdir(d):
                os.makedirs(d)

        self.executable_filename = os.path.join(self.script_dir, "job.sh")
        self.ampel_job_files: dict[int, str] = dict()

        self.manual_submit = manual_submit
        self.node_memory = node_memory
        self.max_materialized_jobs = max_materialized_jobs

        self.job_id: str | None = None
        self._status_output = None

    def make_ampel_job_files(
            self,
            chunks: int | list[int],
    ):
        for c in np.atleast_1d(chunks):
            fn = make_ampel_job_file(base_name=self.base_name, chunk=c, concurrent=True, **self.ampel_options)
            self.ampel_job_files[c] = fn

    def make_executable_file(self):
        """
        Produces the executable that will be submitted to the NPX cluster.
        """

        txt = (
            f"eval \"$({os.environ['CONDA_EXE']} shell.bash hook)\" \n"
            f"conda activate {os.environ['CONDA_PREFIX']} \n"
            f"ssh -Nfl $(whoami) -L localhost:$TIMEWISE_SUP_MONGODB_PORT:localhost:$TIMEWISE_SUP_MONGODB_PORT "
            f"$(whoami)@ztf-wgs.zeuthen.desy.de \n"
            f"ampel job -config {ampel_conf_filename()} -schema $1 -task $2"
        )

        logger.debug("writing executable to " + self.executable_filename)
        with open(self.executable_filename, "w") as f:
            f.write(txt)

    def condor_submit_filename(self):
        return os.path.join(self.script_dir, f"job.submit")

    def make_condor_submit_files(
            self,
            # tasks: list[int],
            # njobs: list[int],
            # chunks: int | list[int],
    ):
        """
        Produces the submit file that will be submitted to the NPX cluster.
        """

        env_str = " ".join([f"{k}={load_environment(k)}" for k in environment_variable_defaults.keys()])
        log_base = os.path.join(self.log_dir, "chunk$(chunk)_task$(task)_$(cluster)_$(process)job")

        text = (
            f"executable = {self.executable_filename} \n"
            f"environment = \"{env_str}\" \n"
            f"log = {log_base}.log \n"
            f"output = {log_base}.out \n"
            f"error = {log_base}.err \n"
            f"should_transfer_files   = YES \n"
            f"when_to_transfer_output = ON_EXIT \n"
            f"arguments = $(ampel_job_file) $(task) \n"
            f"RequestMemory = {self.node_memory} \n"
            f"max_materialized = {self.max_materialized_jobs} \n"
            f"\n"
            f"queue $(njobs)"
        )

        fn = self.condor_submit_filename()
        logger.debug(f"writing submitfile at {fn}")
        with open(fn, "w") as f:
            f.write(text)

    def dagman_filename(
            self,
            chunks: list[int],
            tasks: list[int]
    ):
        _c = "_".join([str(c) for c in chunks])
        _t = "_".join([str(t) for t in tasks])
        _a = ""
        if len(self.ampel_options) > 0:
            for k, v in self.ampel_options.items():
                _a += f"_{k}{v}"

        return os.path.join(self.script_dir, f"chunks{_c}{_a}_tasks{_t}job.dag")

    def make_dagman_file(
            self,
            chunks: list[int],
            tasks: list[int],
            njobs: list[int]
    ):
        txt = ""
        for c in chunks:
            for i, (t, n) in enumerate(zip(tasks, njobs)):
                txt += f"JOB {c}{t} {self.condor_submit_filename()} \n"
                txt += (
                    f"VARS {c}{t} "
                    f"chunk=\"{c}\" "
                    f"ampel_job_file=\"{self.ampel_job_files[c]}\" "
                    f"task=\"{t}\" "
                    f"njobs=\"{n}\"\n"
                )

            for i in range(len(tasks) - 1):
                if i < len(tasks) - 1:
                    txt += f"PARENT {c}{tasks[i]} CHILD {c}{tasks[i+1]}\n"

            txt += "\n\n"

        fn = self.dagman_filename(chunks, tasks)
        logger.debug(f"writing dagfile to {fn}")
        with open(fn, "w") as f:
            f.write(txt)

    def submit_cluster(        
            self,
            tasks: list[int],
            njobs: list[int],
            chunks: int | list[int],
    ):
        """Submits the job to the cluster"""

        _chunks = list(np.atleast_1d(chunks))
        self.make_ampel_job_files(_chunks)
        self.make_executable_file()
        self.make_condor_submit_files()
        self.make_dagman_file(_chunks, tasks, njobs)

        if not self.manual_submit:
            cmd = f"condor_submit_dag {self.dagman_filename(_chunks, tasks)}"
            logger.debug(f"command is {cmd}")
            with subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE) as prc:
                msg = prc.stdout.read().decode()  # type: ignore
            logger.info(msg)

            self.job_id = str(msg).split("cluster ")[-1].split(".")[0]

        else:
            print(
                f"You selected manual submit mode: \n"
                f"\tThe submit file can be found here: \n"
                f"\t{self.dagman_filename(_chunks, tasks)} \n"
                f"\tPlease submit this to the cluster. \n"
            )

    @staticmethod
    def get_condor_status():
        """
        Queries condor to get cluster status.
        :return: str, output of query command
        """
        cmd = "condor_q"
        return subprocess.check_output(cmd).decode()

    def collect_condor_status(self):
        """Gets the condor status and saves it to private attribute"""
        self._status_output = self.get_condor_status()

    @property
    def condor_status(self):
        """
        Get the status of jobs running on condor.
        :return: number of jobs that are done, running, waiting, total, held
        """
        status_list = [
            [y for y in ii.split(" ") if y]
            for ii in self._status_output.split("\n")[4:-6]
        ]
        done = running = waiting = total = held = None

        for li in status_list:
            if li[2] == self.job_id:
                done, running, waiting = li[5:8]
                held = 0 if len(li) == 10 else li[8]
                total = li[-2]

        return done, running, waiting, total, held

    def wait_for_job(self):

        if self.job_id:
            logger.info("waiting for job with ID " + str(self.job_id))
            time.sleep(5)

            self.collect_condor_status()
            j = 0
            while not np.all(np.array(self.condor_status) == None):
                d, r, w, t, h = self.condor_status
                logger.info(
                    f"{time.asctime(time.localtime())} - Job{self.job_id}: "
                    f"{d} done, {r} running, {w} waiting, {h} held of total {t}"
                )
                j += 1
                if j > 7:
                    logger.info(self._status_output)
                    j = 0
                time.sleep(90)
                self.collect_condor_status()

            logger.info("Done waiting for job with ID " + str(self.job_id))

        else:
            logger.info(f"No Job ID!")

    @staticmethod
    def get_pending_ids():
        condor_status = Submitter.get_condor_status()
        ids = np.array([ii.split(" ")[2] for ii in condor_status.split("\n")[4:-6]])
        return ids


def run_bayesian_blocks_on_cluster(
        base_name: str,
        tasks: list[int],
        njobs: list[int],
        chunks: int | list[int],
        node_memory: str,
        manual_submit: bool = False,
        max_materialized_jobs: int = 10,
        **bayesian_blocks_kwargs
):
    logger.info("running bayesian blocks on cluster")
    submitter = Submitter(
        base_name=base_name,
        node_memory=node_memory,
        manual_submit=manual_submit,
        max_materialized_jobs=max_materialized_jobs,
        **bayesian_blocks_kwargs
    )
    logger.debug("submitting to cluster")
    submitter.submit_cluster(chunks=chunks, tasks=tasks, njobs=njobs)
    logger.debug("waiting for job to finish")
    submitter.wait_for_job()



import json
import os.path
from functools import cache
import numpy as np
import logging
import tqdm

import pandas as pd

from timewise_sup.mongo import DatabaseConnector, Index, Status
from timewise_sup.environment import load_environment
from timewise import WiseDataByVisit


logger = logging.getLogger(__name__)


def apply_baseline_subtraction(
        wise_data: WiseDataByVisit,
        database_name: str,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
        status: Status | None = None,
        index: Index | None = None
) -> dict:
    """
    Reads the baseline determined by the T2BayesianBlocks and subtracts that from the datapoints
    """

    logger.debug(f"status {status}")
    logger.debug(f"index {index}")

    database_connector = DatabaseConnector(base_name=wise_data.base_name, database_name=database_name)

    if status is not None:
        logger.debug(f"getting indices from status {status}, ignoring passed value {index}!")
        index = database_connector.get_ids(status=status)
    else:
        if index is None:
            raise ValueError("You must specify either of 'index' or 'status'!")

    logger.info("getting baseline values")

    # keys are stock IDs, values are baseline values
    baselines = database_connector.get_baselines(index=index)
    ids = np.array(list(baselines.keys()))
    logger.debug(f"got baselines for {len(ids)} objects")

    # find the chunk that holds the respective IDs
    chunk_numbers = np.array([wise_data._get_chunk_number(parent_sample_index=s) for s in ids])

    # set up empty directory for baseline corrected lightcurves
    diff_lcs = dict()

    # load the chunks one after the other
    logger.info(f"calculating difference flux for {len(ids)} objects in {len(np.unique(chunk_numbers))} chunks")
    for c in np.unique(chunk_numbers):
        logger.debug(f"chunk {c}")

        if load_from_bigdata_dir or isinstance(wise_data, WiseDataByVisit):
            timewise_data_product = wise_data.load_data_product(
                service=service,
                chunk_number=c,
                use_bigdata_dir=load_from_bigdata_dir
            )
        else:
            timewise_data_product = wise_data.load_data_product(
                service=service,
                chunk_number=c
            )

            # loop through IDs in this chunk
        for s in tqdm.tqdm(ids[chunk_numbers == c], desc=f"apply baseline subtraction to chunk {c}"):
            lc_in = pd.DataFrame.from_dict(timewise_data_product[str(s)]["timewise_lightcurve"])
            lc_out = lc_in[["mean_mjd"]].copy()

            for b in ["W1", "W2"]:

                # loop over flux and flux errors
                f_key = WiseDataByVisit.mean_key + WiseDataByVisit.flux_density_key_ext
                ferr_key = WiseDataByVisit.flux_density_key_ext + WiseDataByVisit.rms_key

                try:
                    # f is a dict with keys: index and values: flux_densities
                    f = lc_in[b + f_key]
                    baseline = baselines[s][f"{b}_baseline"]
                    baseline_err = baselines[s][f"{b}_baseline_sigma"]

                    if baseline is None:
                        logger.warning(f"baseline for {s} {b} is None, skipping!")
                        continue

                    f_diff = {k: v - baseline for k, v in f.items()}
                    f_diff_key = "_diff" + f_key
                    lc_out[b + f_diff_key] = f_diff

                    ferr_diff_key = "_diff" + ferr_key
                    lc_out[b + ferr_diff_key] = np.sqrt(lc_in[b + ferr_key].copy() ** 2 + baseline_err ** 2)

                except KeyError as e:
                    logger.error(
                        f"parent sample index {s}: {e}. "
                        f"LC is \n{json.dumps(lc_out, indent=4)}"
                    )
                    raise KeyError

            # save lightcurve containing baseline corrected flux
            diff_lcs[str(s)] = lc_out.to_dict()

    logger.info("done")

    return diff_lcs


def get_baseline_subtracted_lightcurves(
        base_name: str,
        database_name: str,
        wise_data: WiseDataByVisit,
        status: Status,
        force_new: bool = False,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
) -> dict:
    logger.info(f"getting baseline subtracted lightcurves for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"diff_lcs_status{status}.json")

    if (not os.path.isfile(fn)) or force_new:
        logger.info(f"No file {fn}.")
        logger.info("Making baseline subtracted lightcurves")
        diff_lcs = apply_baseline_subtraction(
            wise_data=wise_data,
            database_name=database_name,
            service=service,
            load_from_bigdata_dir=load_from_bigdata_dir,
            status=status
        )

        d = os.path.dirname(fn)
        if not os.path.isdir(d):
            os.makedirs(d)

        logger.debug(f"saving under {fn}")
        with open(fn, "w") as f:
            json.dump(diff_lcs, f)

    else:
        logger.debug(f"loading {fn}")
        with open(fn, "r") as f:
            diff_lcs = json.load(f)

    return diff_lcs


def get_lightcurves(
        base_name: str,
        database_name: str,
        wise_data: WiseDataByVisit,
        index: Index,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
) -> dict:
    indices = list(np.atleast_1d(index))
    logger.info(f"getting lightcurves {len(indices)} objects")
    status = DatabaseConnector(base_name=base_name, database_name=database_name).get_status(index=tuple(indices))
    logger.debug(f"found {len(status['status'].unique())} statuses")
    lcs = dict()
    for s in status["status"].unique():
        iids = status.index[status["status"] == s]
        slcs = get_baseline_subtracted_lightcurves(
            base_name=base_name,
            database_name=database_name,
            wise_data=wise_data,
            status=s,
            service=service,
            load_from_bigdata_dir=load_from_bigdata_dir
        )

        for i in iids:
            lcs[str(i)] = slcs[i]

    logger.debug(f"returning {len(lcs)} lightcurves")
    return lcs


@cache
def get_single_lightcurve(
        base_name: str,
        database_name: str,
        wise_data: WiseDataByVisit,
        index: str,
        service: str = "tap",
        load_from_bigdata_dir: bool = True,
) -> pd.DataFrame:
    lcs = get_lightcurves(
        base_name=base_name,
        database_name=database_name,
        wise_data=wise_data,
        index=index,
        service=service,
        load_from_bigdata_dir=load_from_bigdata_dir
    )
    return pd.DataFrame.from_dict(lcs[str(index)])



import logging
import os
import json
import pandas as pd
import numpy as np
from pymongo import MongoClient, collection, database
from collections.abc import Sequence
from pydantic import BaseModel
from pathlib import Path

from timewise_sup.environment import load_environment


logger = logging.getLogger(__name__)


database_names = {
    # The sample including NED LVS has the same database name becaus eit's just an
    # emendment to the NEWS + PS1xWISE-STRM sample
    "news_plus_ps1xwisestrm_plus_nedlvs_sample": "news_plus_ps1xwisestrm_sample",
    "news_plus_ps1xwisestrm_sample": "news_plus_ps1xwisestrm_sample",
    "news_sample": "wise_TDE_all_passed_data",
    "ztf_nuclear_summary": "ztfnuclear",
    "ztf_nuclear": "wise_nuclear_data",
    "sjoerts_flares": "sjoert_sample",
    "bts_sample": "bts_sample",
    "ZTF19adgzidh": "ZTF19adgzidh"
}


jd_offset = -2400000.5

################################################################################
# -----------------------         utilities            ----------------------- #
# --------------------------------  START  ----------------------------------- #
#


Index = Sequence[str | int] | str | int
Status = str | Sequence[str]


def as_list(
        index: Index
) -> list:

    indices = list(np.atleast_1d(index))

    for i, ii in enumerate(indices):
        if isinstance(ii, np.int64):
            indices[i] = int(ii)

    return indices


def chunks(
        the_list: Sequence,
        length: int
):
    """Yield sequential chunks from l with fixed length"""
    number_of_chunks = int(np.ceil(len(the_list) / length))
    logger.debug(f"splitting {len(the_list)} into {number_of_chunks} chunks of length {length}")
    d, r = divmod(len(the_list), number_of_chunks)
    for i in range(number_of_chunks):
        logger.debug(f"chunk {i}")
        si = (d+1)*(i if i < r else r) + d*(0 if i < r else i - r)
        yield the_list[si:si+(d+1 if i < r else d)]


#
# -----------------------         utilities            ----------------------- #
# --------------------------------   END   ----------------------------------- #
################################################################################


################################################################################
# -----------------------      DatabaseConnector       ----------------------- #
# --------------------------------  START  ----------------------------------- #
#


class DatabaseConnector(BaseModel):

    base_name: str
    database_name: str
    database: database.Database
    t2collection: collection.Collection

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        # get the client, database and t2 collections
        client: MongoClient = DatabaseConnector.connect_to_db()
        _database_name = kwargs["database_name"]

        if _database_name not in client.list_database_names():
            logger.warning(f"No database with name {_database_name} registered in client!")

        _database = client[_database_name]
        kwargs["database"] = _database
        kwargs["t2collection"] = _database["t2"]

        super().__init__(**kwargs)

    @property
    def cache_dir(self) -> Path:
        tsup_data = Path(load_environment("TIMEWISE_SUP_DATA"))
        return tsup_data / self.base_name / self.database_name

    ################################################################################
    # -----------------------  interface with MongoDB      ----------------------- #
    # --------------------------------  START  ----------------------------------- #
    #

    @staticmethod
    def connect_to_db() -> MongoClient:
        mongodb_port = load_environment("TIMEWISE_SUP_MONGODB_PORT")
        logger.debug(f"connecting to MongoDB at {mongodb_port}")
        client: MongoClient = MongoClient(f"mongodb://localhost:{mongodb_port}/")
        logger.debug("connected")
        return client

    def get_dataframe(
            self,
            collection_name: str,
            field_name_lists: list[list[str | int]],
            filter: dict | None = None
    ) -> pd.DataFrame:
        logger.info(f"making dataframe from {field_name_lists} of {self.database_name}.{collection_name}")
        col = self.database[collection_name]  # type: collection.Collection
        filter = dict() if filter is None else filter

        res: dict[str, list] = {f"field{i}": list() for i in range(len(field_name_lists))}
        for entry in col.find(filter):
            for i_field_names, field_names in enumerate(field_name_lists):

                # loop through the list of field names / indices to dig down to the value
                val = entry
                for field_name in field_names:
                    if field_name not in val:
                        val = np.nan
                        break
                    val = val[field_name]

                res[f"field{i_field_names}"].append(val)

        return pd.DataFrame.from_dict(res)

    #
    # -----------------------  interface with MongoDB      ----------------------- #
    # --------------------------------   END   ----------------------------------- #
    ################################################################################

    ################################################################################
    # ------------------     get T2DustEchoEval results       -------------------- #
    # --------------------------------  START  ----------------------------------- #
    #

    def get_status(
            self,
            index: Index
    ) -> pd.DataFrame:
        indices = as_list(index)
        logger.debug(f"getting status for {len(indices)} IDs ({self.base_name} in {self.database_name})")

        columns = {
            "stock": 1,
            "body.status": 1
        }

        i = None
        status = dict()
        for chunked_indices in chunks(indices, int(1e6)):
            dust_echo_filter = {
                "unit": "T2DustEchoEval",
                "code": 0,
                "stock": {"$in": chunked_indices}
            }
            for i, x in enumerate(self.t2collection.find(dust_echo_filter, columns)):
                status[str(x["stock"])] = {"status": x['body'][-1]['status']}

        if len(status) == 0:
            return pd.DataFrame({"status": []})

        return pd.DataFrame.from_dict(status, orient="index")

    def get_unique_statuses(self):
        filter = {
            "unit": "T2DustEchoEval",
            "code": 0
        }
        statuses = self.t2collection.distinct("body.status", filter)
        return statuses

    def calculate_ids(self, status: Status) -> list:
        logger.debug(f"calculating all IDs for status {status} ({self.base_name} in {self.database_name})")

        filter = {
            "unit": "T2DustEchoEval",
            "body.status": {"$in": as_list(status)},
            "code": 0
        }
        # N = col.count_documents(filter=filter)
        # logger.debug(f"found {N} documents")

        columns = {"stock": 1, "body.status": 1}
        stocks = list()
        N_wrong = 0
        for i in self.t2collection.find(filter, columns):
            if i["body"][-1]["status"] not in status:
                N_wrong += 1
            else:
                stocks.append(i["stock"])

        logger.debug(f"found {len(stocks)} stocks, {N_wrong} had matching status from previous run.")
        return stocks

    def get_ids(self, status: Status) -> list:
        logger.debug(f"getting all IDs for status {status} ({self.base_name} in {self.database_name})")

        ids_cache_dir = self.cache_dir / "ids"
        ids_cache_dir.mkdir(parents=True, exist_ok=True)

        stocks = list()

        for istatus in np.unique(status).astype(str):
            fn = ids_cache_dir / f"status{istatus.replace(' ', '_')}.json"

            if not os.path.isfile(fn):
                istocks = self.calculate_ids(istatus)
                logger.debug(f"caching to {fn}")
                with open(fn, "w") as f:
                    json.dump(istocks, f)

            else:
                logger.debug(f"loading cache from {fn}")
                with open(fn, "r") as f:
                    istocks = json.load(f)

                logger.debug(f"got {len(istocks)} ids")

            stocks.extend(istocks)

        return stocks

    def get_excess_mjd(self, index: Index) -> dict:
        """
        Get the excess time for the given index
        :param index:
        :return:
        """
        indices = as_list(index)
        logger.debug(f"getting excess time for {len(indices)} IDs ({self.base_name} in {self.database_name})")

        excess_mjds = dict()

        for chunked_indices in chunks(indices, int(1e6)):
            bayesian_blocks_filter = {
                "unit": "T2BayesianBlocks",
                "stock": {"$in": chunked_indices},
                "code": 0
            }

            for i, x in enumerate(self.t2collection.find(bayesian_blocks_filter)):
                i_excess_mjds = dict()
                for f in ["W1", "W2"]:
                    excess_mags = np.array(x["body"][-1][f"Wise_{f}"]["max_mag_excess_region"])
                    excess_jds = x["body"][-1][f"Wise_{f}"]["jd_excess_regions"]
                    max_excess_mjds = np.array(excess_jds)[np.argmax(excess_mags)] + jd_offset

                    baseline_mjds = np.array(x["body"][-1][f"Wise_{f}"]["jd_baseline_regions"]).flatten() + jd_offset
                    excess_ended = max(baseline_mjds) > max(max_excess_mjds)

                    i_excess_mjds[f"{f}_excess_start_mjd"] = min(max_excess_mjds)
                    i_excess_mjds[f"{f}_excess_end_mjd"] = max(max_excess_mjds)
                    i_excess_mjds[f"{f}_flare_ended"] = excess_ended

                excess_mjds[str(x["stock"])] = i_excess_mjds

        return excess_mjds

    def get_baselines(self, index: Index) -> dict:
        """
        Returns the baseline of
        """
        baseline = dict()

        indices = as_list(index)
        logger.debug(f"reading baseline values for {len(indices)} objects")

        for chunked_indices in chunks(indices, int(5e5)):

            dust_echo_filter = {
                "unit": "T2DustEchoEval",
                "stock": {"$in": chunked_indices},
                "code": 0
            }

            for x in self.t2collection.find(
                    dust_echo_filter,
                    {"stock": 1}
            ):
                i_baseline = dict()
                for y in self.t2collection.find(
                        {"unit": "T2BayesianBlocks", "stock": x["stock"], "code": 0},
                        {f"body.Wise_{b}.baseline{s}": 1 for b in ["W1", "W2"] for s in ["", "_sigma"]},
                        limit=1
                ):
                    for b in ["W1", "W2"]:
                        i_baseline[f"{b}_baseline"] = y["body"][-1][f"Wise_{b}"]["baseline"]
                        i_baseline[f"{b}_baseline_sigma"] = y["body"][-1][f"Wise_{b}"]["baseline_sigma"]

                baseline[str(x["stock"])] = i_baseline

        return baseline

    def calculate_t2_dust_echo_eval_descriptions(self, index: Index) -> pd.DataFrame:
        indices = as_list(index)
        logger.debug(f"getting description for {len(indices)} objects of {self.base_name} in {self.database_name}")

        values = dict()

        for chunked_indices in chunks(indices, int(5e5)):
            filter = {
                "code": 0,
                "unit": "T2DustEchoEval",
                "stock": {"$in": chunked_indices}
            }

            for lc in self.t2collection.find(filter):
                value = dict()

                try:
                    value[f"description"] = ", ".join(lc["body"][-1]["description"])
                except KeyError as e:
                    raise KeyError(f"{lc}: {e}")

                values[str(lc["stock"])] = value

        logger.debug(f"returning {len(values)} results")

        return pd.DataFrame.from_dict(values, orient="index")

    def get_t2_dust_echo_eval_descriptions(self, indices: Index) -> pd.DataFrame:
        fn = self.cache_dir / "T2DustEchoEvalDescriptions.csv"
        fn.parents[0].mkdir(parents=True, exist_ok=True)

        if not os.path.isfile(fn):
            logger.debug(f"no file {fn}")
            desc = self.calculate_t2_dust_echo_eval_descriptions(indices)
            logger.debug(f"saving to {fn}")
            desc.to_csv(fn)

        else:
            logger.debug(f"loading {fn}")
            desc = pd.read_csv(fn, index_col=0)

        indices_series = pd.Series(indices)
        indices_present = indices_series.astype(str).isin(desc.index.astype(str))
        if np.any(~indices_present):
            logger.debug(f"{np.sum(indices_present)} indices not found. calculating")
            desc_suplement = self.calculate_t2_dust_echo_eval_descriptions(indices_series[~indices_present])
            desc = pd.concat([desc, desc_suplement])
            logger.debug(f"saving {len(desc)} descriptions to {fn}")
            desc.to_csv(fn)

        return desc

    def get_agn_variability_stocks(self) -> list:
        logger.debug("getting AGN stock IDs based on variability")

        filter = {
            '$and': [
                {"code": 0},
                {"unit": "T2DustEchoEval"},
                # {"body": {"$elemMatch": {"description": "Excess region exists"}}},
                {"body.description": {"$elemMatch": {"$ne": "Baseline only"}}},
                {"body.description": {"$elemMatch": {"$ne": "Stage transition"}}},
                {"body.status": {"$ne": "1"}}
                ]
        }

        cols = {"stock": 1}

        stocks = list()
        for lc in self.t2collection.find(filter, cols):
            stocks.append(lc["stock"])

        logger.debug(f"returning {len(stocks)} stock IDs")
        return stocks

    def get_t2_dust_echo_eval_values(self, index: Index, value: str) -> pd.DataFrame:

        indices = as_list(index)
        logger.debug(f"getting {value} for {len(indices)} objects of {self.base_name} in {self.database_name}")

        filter = {
            "unit": "T2DustEchoEval",
            "stock": {"$in": as_list(index)},
            "code": 0
        }

        n = self.t2collection.count_documents(filter)
        logger.debug(f"found {n} documents")

        values = dict()

        for lc in self.t2collection.find(filter):
            ivalue = dict()

            for i, b in enumerate(["W1", "W2"]):
                try:
                    ivalue[f"{value}_{b}"] = lc["body"][-1]["values"][value][i]
                except KeyError as e:
                    raise KeyError(f"{lc}: {e}")
                except IndexError as e:
                    raise IndexError(f"{lc}: {e}")

            values[str(lc["stock"])] = ivalue

        logger.debug(f"returning {len(values)} results")

        return pd.DataFrame.from_dict(values, orient="index")

    def get_peak_time_jd(self, index: Index) -> pd.DataFrame:
        return self.get_t2_dust_echo_eval_values(index, "max_mag_jd")

    def get_excess_start_jd(self, index: Index) -> pd.DataFrame:
        return self.get_t2_dust_echo_eval_values(index, "excess_jd")

    def get_dust_echo_strength(self, index: Index) -> pd.DataFrame:
        return self.get_t2_dust_echo_eval_values(index, "strength_sjoert")

    def get_fade_time(self, index: Index) -> pd.DataFrame:
        return self.get_t2_dust_echo_eval_values(index, "e_fade")

    def get_rise_time(self, index: Index) -> pd.DataFrame:
        return self.get_t2_dust_echo_eval_values(index, "e_rise")

    #
    # ------------------     get T2DustEchoEval results       -------------------- #
    # --------------------------------   END   ----------------------------------- #
    ################################################################################

    ############################################################################################
    # -------------     get T2CatalogMatch and T2DigestRedshifts results       --------------- #
    # -------------------------------------  START  ------------------------------------------ #
    #

    def get_catalog_matches(self, index: Index) -> dict:
        logger.debug(f"getting catalog match info for {self. base_name} in {self.database_name}")
        indices = as_list(index)
        logger.debug(f"getting catalog match info for {len(indices)} objects")

        matches = dict()

        for chunked_indices in chunks(indices, 1000000):

            filter = {
                "code": 0,
                "stock": {"$in": chunked_indices},
                "unit": "T2CatalogMatch"
            }

            for i in self.t2collection.find(filter):
                matches[i["stock"]] = i["body"][-1]

        return matches

    def get_redshift(self, index: Index) -> pd.DataFrame:
        filter = {
            "unit": "T2DigestRedshifts",
            "stock": {"$in": as_list(index)},
            "body.ampel_z": {"$exists": True}
        }

        n = self.t2collection.count_documents(filter)
        logger.debug(f"found {n} documents")

        redshift = dict()

        for lc in self.t2collection.find(filter):
            try:
                d = {k: lc["body"][-1][k]
                     for k in ["ampel_z", "ampel_dist", "group_z_precision"]}
                redshift[str(lc["stock"])] = d
            except KeyError as e:
                raise KeyError(f"{lc}: {e}")

        logger.debug(f"returning {len(redshift)} results")

        return pd.DataFrame.from_dict(redshift, orient="index")

    #
    # -------------     get T2CatalgueMatch and T2DigestRedshifts results       --------------- #
    # --------------------------------------   END   ------------------------------------------ #
    #############################################################################################

#
# -----------------------      DatabaseConnector       ----------------------- #
# --------------------------------   END   ----------------------------------- #
################################################################################



import logging
import json
import numpy as np
import pandas as pd

from timewise_sup.mongo import DatabaseConnector


logger = logging.getLogger(__name__)


def get_rejection_reason(
        base_name: str,
        database_name: str
) -> dict:
    logger.info(f"getting rejection reason for {base_name}")
    rejected = (
        "No further investigation",
        "1_maybe_interesting",
        "2_maybe_interesting",
        "3", "3_maybe_interesting",
        "4", "4_maybe_interesting"
    )
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(rejected)
    desc = database_connector.get_t2_dust_echo_eval_descriptions(ids)
    desc_counts = desc.description.value_counts()  # type: pd.Series

    individual_reasons = dict()
    for ir, n in desc_counts.items():
        for irsi in ir.split(", "):
            if irsi not in individual_reasons:
                individual_reasons[irsi] = n
            else:
                individual_reasons[irsi] += n

    # sort by number
    individual_reasons_df = pd.DataFrame.from_dict(individual_reasons, orient="index").sort_values(0, ascending=False)

    # find individual reasons in descriptions
    accept = ['Baseline before excess region', 'Excess reg'
                                               ''
                                               ''
                                               ''
                                               ''
                                               'ion exists']
    masks: dict[str, bool] = {
        rt: desc_counts.index.str.contains(rt)
        for rt in individual_reasons_df.index if rt not in accept
    }

    # find individual reasons that exclude the object with descending order of appearance
    masks_exclusive = {rt: im & ~np.logical_or.reduce(list(masks.values())[:i])
                       for i, (rt, im) in enumerate(masks.items())}

    # find the corresponding numbers of objects
    numbers_exclusive = {rt: int(desc_counts[m].sum()) for rt, m in masks_exclusive.items()}
    logger.debug(json.dumps(numbers_exclusive, indent=4))

    return numbers_exclusive



import logging
import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.time import Time
from timewise.wise_data_base import WISEDataBase

from timewise_sup.plots import plots_dir
from timewise_sup.meta_analysis.diagnostics import (
    get_statuses_per_chunk,
    get_catalog_matches_per_chunk,
    get_positional_outliers_times
)


logger = logging.getLogger(__name__)


def chunk_distribution_plots(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase
):
    logger.info(f"making chunk distribution diagnostic plots for {base_name}")
    d = plots_dir("diagnostics", base_name)

    # --- plot analyses stocks per chunk ---#

    statuses = get_statuses_per_chunk(base_name, database_name, wise_data)

    logger.info("plotting number of analysed objects")
    n_stocks = {c: len(v) for c, v in statuses.items()}
    fig, ax = plt.subplots()
    ax.bar(np.array(list(n_stocks.keys())).astype(int), n_stocks.values())
    ax.set_xlabel("chunk number")
    ax.set_ylabel("# of analysed objects")
    for loc in ["top", "right"]:
        ax.spines[loc].set_visible(False)

    ax.grid(ls=":", alpha=0.5)

    fig.tight_layout()
    fn = os.path.join(d, "number_per_chunk.pdf")
    logger.debug(f"saving under {fn}")
    fig.savefig(fn)
    plt.close()

    # --- plot distribution of statuses among chunks ---#

    logger.info("plotting status distribution among chunks")
    _unique_statuses = [
        "1",
        "1_maybe_interesting",
        "2",
        "2_maybe_interesting",
        "3",
        "3_maybe_interesting",
        "4",
        "4_maybe_interesting",
        "No further investigation"
    ]

    n_status = dict()
    for s in _unique_statuses:
        logger.debug(f"getting # of objects for {s}")
        in_status = dict()
        for c, istatus in statuses.items():
            in_status[c] = np.sum(np.array(istatus) == s)

        n_status[s] = in_status

    for s, in_status in n_status.items():
        logger.debug(f"plotting {s}")
        fig, ax = plt.subplots()
        ax.bar(np.array(list(in_status.keys())).astype(int), list(in_status.values()))
        ax.grid(ls=":", alpha=0.5)
        ax.set_xlabel("chunk number")
        ax.set_ylabel("# of objects")
        ax.set_title(s)
        for loc in ["top", "right"]:
            ax.spines[loc].set_visible(False)

        fig.tight_layout()
        fn = os.path.join(d, f"number_per_chunk_status_{s}.pdf")
        logger.debug(f"saving under {fn}")
        fig.savefig(fn)
        plt.close()

    # --- plot distribution of catalogue matches among chunks --- #

    number_of_catalogue_matches = pd.DataFrame.from_dict(
        get_catalog_matches_per_chunk(base_name, database_name, wise_data),
        orient="index"
    ).fillna(0)

    fig, ax = plt.subplots()

    bottom = None
    for i, c in enumerate(number_of_catalogue_matches.columns):
        bottom = (
            0 if i == 0 else
            bottom + number_of_catalogue_matches[number_of_catalogue_matches.columns[i - 1]]
        )
        ax.bar(
            number_of_catalogue_matches.index,
            number_of_catalogue_matches[c],
            bottom=bottom,
            label=c
        )

    for loc in ["top", "right"]:
        ax.spines[loc].set_visible(False)

    ax.grid(ls=":", alpha=0.5)
    ax.legend()
    ax.set_xlabel("chunk number")
    ax.set_ylabel("number of matches")

    fig.tight_layout()
    fn = os.path.join(d, f"number_of_catalog_matches_per_chunk.pdf")
    logger.debug(f"saving under {fn}")
    fig.savefig(fn)
    plt.close()


def positional_outliers(
        base_name: str,
        wise_data: WISEDataBase
):
    logger.info("plotting times of positional outliers")
    mjds_per_chunk = get_positional_outliers_times(base_name, wise_data)

    ic = 0.9

    data = np.array([
        [c] + list(np.quantile(mjds, [0.5, 0.5 - ic / 2, 0.5 + ic / 2]))
        for c, mjds in mjds_per_chunk.items()
    ])

    # -------- make plot --------- #

    fig, ax = plt.subplots()
    color = "r"
    lw = 2
    ax.plot(data[:, 0], data[:, 1], color=color, ls="-", label="median", lw=lw)
    ax.plot(data[:, 0], data[:, 2], color=color, ls="--", label="IC$_{" + f"{ic*100:.0f}" + "}%$", lw=lw)
    ax.plot(data[:, 0], data[:, 3], color=color, ls="--", lw=lw)
    ax.axhline(Time("2011-02-01").mjd, ls=":", label="WISE decommissioned")
    ax.axhline(Time("2013-09-01").mjd, ls=":", label="WISE reactivated")

    ax.set_xlabel("chunk number")
    ax.set_ylabel("MJD")
    ax.grid(ls=":", alpha=0.5)

    fig.tight_layout()
    fn = os.path.join(plots_dir("diagnostics", base_name), f"positional_outliers_times.pdf")
    logger.info(f"saving under {fn}")
    fig.savefig(fn)
    plt.close()



import logging
import os
from collections.abc import Sequence
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from timewise.wise_data_base import WISEDataBase

from timewise_sup.plots import plots_dir
from timewise_sup.meta_analysis.ir_energy import get_ir_energy_status
from timewise_sup.mongo import Status


logger = logging.getLogger(__name__)


def ir_energy_hist(
        base_name: str,
        database_name: str,
        status: Status,
        wise_data: WISEDataBase | None = None,
):
    logger.info("plotting bolometric luminosities histogram")

    if isinstance(status, str):
        status = [status]

    fig, ax = plt.subplots()

    lums = list()
    for s in status:
        df = pd.DataFrame.from_dict(get_ir_energy_status(base_name, database_name, s, wise_data), orient="index")
        df["status"] = s
        lums.append(df)

    lums_df = pd.concat(lums)
    pos_mask = lums_df["ir_energy_erg"] > 0

    sns.histplot(
        data=lums_df[pos_mask],
        x="ir_energy_erg",
        hue="status",
        log_scale=True,
        ax=ax,
        kde=True,
        stat="density",
        common_norm=False,
    )

    ax.set_xlabel("IR energy [erg]")
    ax.set_ylabel("density")
    ax.grid(ls=":", zorder=10, alpha=0.5)

    for loc in ["top", "right"]:
        ax.spines[loc].set_visible(False)

    fn = os.path.join(plots_dir("ir_energy", base_name), "ir_energy_hist.pdf")
    logger.debug(f"writing {fn}")
    fig.savefig(fn, bbox_inches="tight")
    plt.close()



import logging
import os.path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import corner
import pandas as pd
from timewise import WISEDataDESYCluster

from timewise_sup.plots import plots_dir
from timewise_sup.meta_analysis.agn import add_agn_box, get_agn_mask_from_color
from timewise_sup.mongo import DatabaseConnector, Index, Status


logger = logging.getLogger(__name__)


def color_plot(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster,
        color_keys: list[str],
        status: Status | None = None,
        colors: list[str] | None = None,
        level_sigma: list[float] | None = None,
        data: pd.DataFrame | None = None,
) -> tuple[mpl.figure.Figure, mpl.axes.Axes]:
    """

    """

    if status is None:
        status = ["1", "2"]

    if colors is None:
        colors = ["k", "grey"]

    if level_sigma is None:
        level_sigma = [1, 2, 5]

    if data is None:
        data = wise_data.parent_sample.df
        data.set_index(data.index.astype(str), inplace=True)

    fig, axs = plt.subplots(
        ncols=len(color_keys),
        nrows=len(color_keys),
        figsize=[6, 6],
        sharex='all',
        sharey='all',
        gridspec_kw={"wspace": 0.1, "hspace": 0.1}
    )

    twinx = [None] * len(color_keys)

    for i_status, c in zip(status, colors):

        ids = DatabaseConnector(base_name=base_name, database_name=database_name).get_ids(i_status)

        for i in range(len(color_keys)):
            for j in range(i + 1, len(color_keys)):
                x = np.array(data.loc[list(ids), color_keys[i]])
                y = np.array(data.loc[list(ids), color_keys[j]])

                logger.debug(f"making corner plot: {color_keys[i]} ({len(x)}) - {color_keys[j]} ({len(y)})")
                nan_m = np.isnan(x) | np.isnan(y)
                logger.debug(f"removing {np.sum(nan_m)} NaNs")

                corner.hist2d(
                    x[~nan_m],
                    y[~nan_m],
                    ax=axs[j][i],
                    plot_density=False,
                    color="k",
                    data_kwargs={"color": c},
                    contour_kwargs={"colors": c},
                    levels=1.0 - np.exp(-0.5 * np.array(level_sigma) ** 2)
                )

                axs[i][j].axis('off')

            if twinx[i] is None:
                twinx[i] = axs[i][i].twinx()

            twinx[i].hist(  # type: ignore
                data.loc[ids, color_keys[i]],
                color=c,
                density=True,
                histtype='step',
                bins=100
            )

            axs[i][0].set_ylabel(color_keys[i])
            axs[-1][i].set_xlabel(color_keys[i])

    return fig, axs


def agn_box_plot(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster,
        status: Status | None = None,
        colors: list[str] | None = None,
        level_sigma: list[float] | None = None,
):
    """
    Makes a color-color plot in the W1-W2 - W2-W3 plane and indicates the `AGN-Wedge`
    (Hvding et al. (2022) https://iopscience.iop.org/article/10.3847/1538-3881/ac5e33).
    """

    logger.info("Making AGN-box color plot")

    if status is None:
        status = ["1", "2"]

    if colors is None:
        colors = ["g", "purple"]

    if level_sigma is None:
        level_sigma = [1, 5]

    parent_sample = wise_data.parent_sample
    wise_mag_keys = ["W1_mag", "W2_mag", "W3_mag"]

    if np.any([k not in parent_sample.default_keymap for k in wise_mag_keys]):
        logger.warning(f"One of {','.join(wise_mag_keys)} not in ParentSample's default keymap: \n"
                       f"{parent_sample.default_keymap}.\n"
                       f"Not making AGN-Box plot.")
        return

    agn_mask, W1W2, W2W3 = get_agn_mask_from_color(base_name, wise_data)
    parent_sample.df["W1-W2"] = W1W2
    parent_sample.df["W2-W3"] = W2W3

    n_agn = dict()

    for istatus in status:
        ids = DatabaseConnector(base_name=base_name, database_name=database_name).get_ids(istatus)
        status_agns = agn_mask.loc[ids]
        in_agns = np.sum(status_agns) / len(status_agns)
        logger.info(f"Status {istatus}: {in_agns*100:.2f}% in AGN box")
        n_agn[istatus] = in_agns

    fig, axs = color_plot(
        base_name,
        database_name,
        wise_data,
        status=status,
        colors=colors,
        color_keys=['W2-W3', 'W1-W2'],
        level_sigma=level_sigma,
        data=parent_sample.df
    )

    axs[0][0].set_xlim(-1, 7)
    axs[0][0].set_ylim(-1, 3)

    add_agn_box(axs[1][0])
    axs[1][0].legend()

    lines = [mpl.lines.Line2D([0], [0], color=c) for c in colors]
    labels = [f"{s} ({n_agn[s] * 100:.0f}% in AGN box)" for s in status]
    fig.legend(lines, labels, loc='lower left', fontsize="medium", bbox_to_anchor=(0.56, 0.56))

    fn = os.path.join(plots_dir("color_plots", base_name), "agn_box.pdf")
    logger.debug(f"saving under {fn}")
    fig.savefig(fn)

    plt.close()



import logging
import os
import json
import numpy as np
from numpy import typing as npt
import pandas as pd
from astropy import units as u
from astropy.cosmology import Planck18
from astropy import constants
from timewise.wise_data_base import WISEDataBase

from timewise_sup.environment import load_environment
from timewise_sup.mongo import DatabaseConnector, Index, Status
from timewise_sup.meta_analysis.baseline_subtraction import get_lightcurves


logger = logging.getLogger(__name__)


band_wavelengths = {
    "W1": 3.4 * 1e-6 * u.m,   # from Wright et al. (2010) ( 10.1088/0004-6256/140/6/1868 )
    "W2": 4.6 * 1e-6 * u.m,
    "ZTF_g": 4804.79 * u.AA,  # from http://svo2.cab.inta-csic.es/svo/theory/fps3/index.php?id=Palomar/ZTF.g&&mode=browse&gname=Palomar&gname2=ZTF
    "ZTF_r": 6436.92 * u.AA,
    "ZTF_i": 7968.22 * u.AA
}


def get_band_nu(band: str) -> float:
    wl = band_wavelengths[band]
    return constants.c / wl


def nuFnu(
        flux: list[float],
        flux_unit: str,
        band: str,
        out_unit: str = "erg s-1 cm-2"
) -> npt.NDArray:
    _flux = np.array(flux) * u.Unit(flux_unit)
    nu = get_band_nu(band)
    return np.array(u.Quantity(_flux * nu).to(out_unit).value)


def calculate_ir_luminosities(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        redshift: dict[str, float] | pd.Series,
        redshift_err: dict[str, float] | pd.Series,
) -> dict:
    indices = tuple(np.atleast_1d(index).astype(int))
    logger.info(f"calculating luminosities for {len(indices)} objects")

    if len(indices) != len(redshift):
        raise ValueError("redshift and index must have the same length!")

    lcs = get_lightcurves(base_name, database_name, wise_data, indices)
    logger.debug(f"got {len(lcs)} lightcurves")
    lcs_with_luminosities = dict()
    for i, lc_dict in lcs.items():
        lc = pd.DataFrame.from_dict(lc_dict, orient="columns")
        iredshift = redshift[i]
        iredshift_err = redshift_err[i]

        lum_dist = Planck18.luminosity_distance(iredshift)
        area = 4 * np.pi * lum_dist ** 2

        lum_dist_ic = Planck18.luminosity_distance(iredshift + iredshift_err * np.array([-1, 1]))
        area_ic = 4 * np.pi * lum_dist_ic ** 2

        for b in ["W1", "W2"]:
            lc[f"{b}_nuFnu_erg_per_s_per_sqcm"] = nuFnu(
                lc[f"{b}_diff_mean_flux_density"],
                flux_unit="mJy",
                band=b,
                out_unit="erg s-1 cm-2"
            )

            lc[f"{b}_nuFnu_err_erg_per_s_per_sqcm"] = nuFnu(
                lc[f"{b}_diff_flux_density_rms"],
                flux_unit="mJy",
                band=b,
                out_unit="erg s-1 cm-2"
            )

            nuFnu_val = u.Quantity(lc[f"{b}_nuFnu_erg_per_s_per_sqcm"] * u.Unit("erg s-1 cm-2"))
            nuFnu_valerr = u.Quantity(lc[f"{b}_nuFnu_err_erg_per_s_per_sqcm"] * u.Unit("erg s-1 cm-2"))

            lum = u.Quantity(nuFnu_val * area).to("erg s-1").value
            lum_err = u.Quantity(
                np.sqrt(
                    (nuFnu_valerr * area) ** 2 + (nuFnu_val * max(abs(area-area_ic))) ** 2
                )
            ).to("erg s-1").value

            lc[f"{b}_ir_luminosity_erg_per_s"] = lum
            lc[f"{b}_ir_luminosity_err_erg_per_s"] = lum_err

        lcs_with_luminosities[i] = lc.to_dict()

    return lcs_with_luminosities


def get_ir_luminosities(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: Status,
        force_new: bool = False
) -> dict:

    logger.info(f"getting luminosities for status {status}")
    tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")
    fn = os.path.join(tsup_data_dir, base_name, f"lum_ir_lcs_status{status}.json")

    if (not os.path.isfile(fn)) or force_new:
        logger.debug(f"No file {fn}")
        logger.info("calculating luminosities")

        database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
        indices = database_connector.get_ids(tuple(status))
        redshifts = database_connector.get_redshift(indices)
        lcs = calculate_ir_luminosities(
            base_name,
            database_name,
            wise_data,
            redshifts.index,
            redshifts.ampel_z,
            redshifts.group_z_precision
        )

        logger.debug(f"writing to {fn}")
        with open(fn, "w") as f:
            json.dump(lcs, f)

    else:
        logger.debug(f"reading from {fn}")
        with open(fn, "r") as f:
            lcs = json.load(f)

    return lcs


# def get_peak_ir_luminosity_status(
#         base_name: str,
#         database_name: str,
#         status: (list[str], str),
# ) -> dict:
#     logger.info(f"getting peak IR luminosities for status {status} ({base_name})")
#     lcs = get_ir_luminosities(base_name, database_name, status)
#     logger.debug(f"got {len(lcs)} lightcurves")
#
#     peak_lum = dict()
#     for i, lc_dict in lcs.items():
#         lc = pd.DataFrame.from_dict(lc_dict, orient="columns")
#         peak_lum[i] = dict()
#         for b in ["W1", "W2"]:
#             arg = np.argmax(lc[f"{b}_luminosity_erg_per_s"])
#             peak_lum[i][f"{b}_peak_luminosity_erg_per_s"] = lc[f"{b}_luminosity_erg_per_s"][arg]
#             peak_lum[i][f"{b}_peak_luminosity_err_erg_per_s"] = lc[f"{b}_luminosity_err_erg_per_s"][arg]
#             peak_lum[i][f"{b}_peak_luminosity_mjd"] = lc["mean_mjd"][arg]




import logging
import json
from argparse import ArgumentParser

from timewise_sup.config_loader import TimewiseSUPConfigLoader


logger = logging.getLogger(__name__)


def main():
    parser = ArgumentParser()
    parser.add_argument("config", type=str, help="Path to timewise_sup config file")
    parser.add_argument("-l", "--logging-level", default="INFO", type=str)
    cfg = vars(parser.parse_args())

    logging_level = cfg.pop("logging_level")
    logging.getLogger("timewise_sup").setLevel(logging_level)
    logging.getLogger("timewise").setLevel(logging_level)
    logger.debug(f"Running timewise_sup with args {json.dumps(cfg, indent=4)}")
    TimewiseSUPConfigLoader.run_yaml(cfg["config"])



import os
import logging
import subprocess

from timewise_sup.environment import load_environment


logger = logging.getLogger(__name__)


def ampel_conf_filename():
    tsup_data = load_environment("TIMEWISE_SUP_DATA")
    return os.path.join(tsup_data, "ampel_config.yml")


def create_ampel_config_file():

    conf_filename = ampel_conf_filename()

    dir = os.path.dirname(conf_filename)
    if not os.path.isdir(dir):
        os.makedirs(dir)

    cmd = ["ampel", "config", "build", "-out", conf_filename]
    logger.debug(f"executing {' '.join(cmd)}")

    with subprocess.Popen(cmd) as p:
        out, err = p.communicate()
        if out is not None:
            logger.info(out.decode())
        if err is not None:
            logger.error(err.decode())

    logger.debug("manually replacing MongoDB port")

    with open(conf_filename, "r") as f:
        content = f.read()
        new_content = content.replace(
            'mongodb://localhost:27017',
            'mongodb://localhost:' + load_environment('TIMEWISE_SUP_MONGODB_PORT')
        )

    with open(conf_filename, "w") as f:
        f.write(new_content)


def get_catalogue_match_conf(match_dist: float) -> dict:
    cat_conf = {
        "catalogs": {
            "SDSS_spec": {"use": "extcats", "rs_arcsec": match_dist, "keys_to_append": None},
            "NEDz": {
                "use": "catsHTM",
                "rs_arcsec": match_dist,
                "keys_to_append": ["ObjType", "Velocity", "z"],
            },
            "NEDz_extcats": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": ["ObjType", "Velocity", "z"],
            },
            "GLADEv23": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": ["z", "dist", "dist_err", "flag1", "flag2", "flag3"],
            },
            "LSPhotoZZou": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": [
                    "photoz",
                    "ra",
                    "dec",
                    "e_photoz",
                    "specz",
                    "_6",
                    "logMassBest",
                    "logMassInf",
                    "logMassSup",
                ],
                "pre_filter": None,
                "post_filter": None,
                "all": False,
            },
            "wiseScosPhotoz": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": [
                    "zPhoto_Corr",
                    "ra",
                    "dec",
                    "wiseID",
                    "w1mCorr",
                    "w2mCorr",
                ],
                "pre_filter": None,
                "post_filter": None,
            },
            "twoMPZ": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": ["zPhoto", "ra", "dec", "zSpec"],
                "pre_filter": None,
                "post_filter": None,
            },
            "TNS": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": None,
            },
            "milliquas": {
                "use": "extcats",
                "rs_arcsec": match_dist,
                "keys_to_append": ["broad_type", "name", "redshift", "qso_prob"],
                "pre_filter": None,
                "post_filter": None,
            },
        }
    }
    return cat_conf


def get_local_catalogue_match_conf(match_dist: float) -> dict:
    cat_local_conf = {
        'catalogs': {
            'fermi4LAC': {
                'use': 'extcats',
                'rs_arcsec': match_dist,
                'keys_to_append': ['Source_Name', 'ra', 'dec'],
                'pre_filter': None,
                'post_filter': None,
                'all': False,
            },
            'BROS': {
                'use': 'extcats',
                'rs_arcsec': match_dist,
                'keys_to_append': ['Source_Name', 'ra', 'dec'],
                'pre_filter': None,
                'post_filter': None,
                'all': False,
            },
            'crates': {
                'use': 'extcats',
                'rs_arcsec': match_dist,
                'keys_to_append': ['Source_Name', 'ra', 'dec'],
                'pre_filter': None,
                'post_filter': None,
                'all': False,
            },
            'CGRaBS': {
                'use': 'extcats',
                'rs_arcsec': match_dist,
                'keys_to_append': ['Source_Name', 'ra', 'dec'],
                'pre_filter': None,
                'post_filter': None,
                'all': False,
            },
            'ROMA_BZCAT': {
                'use': 'extcats',
                'rs_arcsec': match_dist,
                'keys_to_append': ['Source_Name', 'ra', 'dec'],
                'pre_filter': None,
                'post_filter': None,
                'all': False,
            },
            '3HSP': {
                'use': 'extcats',
                'rs_arcsec': match_dist,
                'keys_to_append': ['Source_Name', 'ra', 'dec'],
                'pre_filter': None,
                'post_filter': None,
                'all': False,
            },
        },
        'closest_match': True
    }
    return cat_local_conf


def get_bandpass_filter_component(fid: int) -> dict:
    return {"attribute": "fid", "value": fid, "operator": "=="}


def get_filter_config(logical_connection, red_chi2_threshold: float | None = None):

    if red_chi2_threshold is not None:
        red_chi2_filter = {"attribute": "red_chi2", "value": red_chi2_threshold, "operator": ">"}
        criterias = [[get_bandpass_filter_component(fid), red_chi2_filter] for fid in [1, 2]]

    else:
        # mypy can not handle list comprehensions well
        criterias = [get_bandpass_filter_component(fid) for fid in [1, 2]]  # type: ignore

    filters = [{"criteria": criterias[0], "len": 1, "operator": ">="},
               {"logicalConnection": logical_connection, "criteria": criterias[1], "len": 1, "operator": ">="}]

    filter_config = {"unit": "BasicMultiFilter", "config": {"filters": filters}}
    return filter_config


import os
import json
import logging
import pandas as pd
import numpy as np
from astropy.time import Time
from tqdm import tqdm
from datetime import datetime
import pytz
from timewise.wise_data_base import WISEDataBase
try:
    from fpbot.pipeline import ForcedPhotometryPipeline, FORCEPHOTODATA
except ImportError:
    ForcedPhotometryPipeline = FORCEPHOTODATA = None

from timewise_sup.mongo import DatabaseConnector, Index, Status


logger = logging.getLogger(__name__)

ztf_start = Time(2458194.5, format="jd")  # from https://irsa.ipac.caltech.edu/data/ZTF/docs/ztf_forced_photometry.pdf
now = Time(datetime.now(pytz.timezone("UTC")))


def check_fpbot():
    if ForcedPhotometryPipeline is None:
        raise ImportError(f"'fpbot' is not installed! Make sure to install timewise-sup with the extra 'fpbot'.")


def download_ztffp(
        base_name: str,
        wise_data: WISEDataBase,
        indices: Index,
        time_intervals: pd.DataFrame,
        download: bool = True,
        psf_fit: bool = True
):
    check_fpbot()
    # select the objects and the data needed for FPBot
    _indices = np.atleast_1d(indices)
    logger.info(f"Downloading ZTF forced photometry for {len(_indices)} objects from {base_name}")
    selection = wise_data.parent_sample.df.loc[np.array(_indices).astype(int)]
    cols = [wise_data.parent_sample.default_keymap[k] for k in ["id", "ra", "dec"]]
    columns = {wise_data.parent_sample.default_keymap[k]: k for k in ["id", "ra", "dec"]}
    renamed_selection = selection[cols].rename(columns=columns)

    pipelines = dict()
    skipped = 0
    logger.info("setting up pipelines")
    for index, r in renamed_selection.iterrows():

        time_interval_jd = time_intervals.loc[str(index), ["t_start", "t_end"]]
        logger.debug(f"{index}: time interval: {time_interval_jd.values}")
        if time_interval_jd[1] < ztf_start.jd:
            logger.debug(f"{index}: t_end={time_interval_jd[0]:.0f} which is before ZTF.")
            skipped += 1

        else:
            p = ForcedPhotometryPipeline(
                file_or_name=f'{base_name}_{index}',
                ra=r["ra"],
                dec=r["dec"],
                jdmin=time_interval_jd[0],
                jdmax=time_interval_jd[1]
            )
            pipelines[index] = p

    logger.info(f"skipped {skipped} of {len(renamed_selection)}: before ZTF.")

    if download:
        logger.info("starting download")
        for index, p in tqdm(pipelines.items(), desc="downloading", total=len(pipelines)):
            p.download()

    if psf_fit:
        logger.info("starting psf fits")
        for index, p in tqdm(pipelines.items(), desc="fitting PSFs", total=len(pipelines)):
            try:
                p.psffit()
            except FileNotFoundError:
                logger.debug(f"No observations for {index}")


def download_ztffp_per_status(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        status: str,
        time_interval_relative_to_peak: tuple[float, float] | None = None,
        download: bool = True,
        psf_fit: bool = True
):
    check_fpbot()
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    ids = database_connector.get_ids(status)
    peak_times = database_connector.get_peak_time_jd(index=ids)
    peak_times["t_ref"] = peak_times.median(axis=1)

    if time_interval_relative_to_peak is None:
        peak_times["t_start"] = ztf_start.jd
        peak_times["t_end"] = now.jd

    else:
        peak_times["t_start"] = peak_times["t_ref"] + time_interval_relative_to_peak[0]
        peak_times["t_end"] = peak_times["t_ref"] + time_interval_relative_to_peak[1]

    download_ztffp(base_name, wise_data, ids, peak_times, download=download, psf_fit=psf_fit)


def get_ztf_lightcurve(
        base_name: str,
        index: str,
        snr_threshold: float = 5,
) -> pd.DataFrame | None:
    check_fpbot()
    fn = os.path.join(FORCEPHOTODATA, f"{base_name}_{index}.csv")

    if not os.path.isfile(fn):
        logger.debug(f"No file for {index} in {base_name}.")
        return None

    data = pd.read_csv(fn, comment="#")

    # calculate the proportionality constant between a.u. and flux density
    c = 10 ** ((8.9 - data["magzp"]) / 2.5)

    # the flux is just the instrument value times the prop. constant
    flux = data["ampl"] * c
    # the error is calculated by gaussian propagation of the instrument value and the
    # error on the magnitude zeropoint (data["magzpunc"])
    flux_error = np.sqrt(
        (data["ampl.err"] * c) ** 2 +
        (data["ampl"] * c * np.log(10) / 2.5 * data["magzpunc"]) ** 2
    )

    data["flux[Jy]"] = flux
    data["flux_err[Jy]"] = flux_error
    data[f"above_snr"] = (flux / flux_error) > snr_threshold

    return data


def get_ztf_id_from_nuclear_sample(
        base_name: str,
        wise_data: WISEDataBase,
        index: str
) -> dict | None:
    # TODO: add upper limits!
    logger.info(f"Getting crossmatch to {index} of {base_name} from ZTF nuclear")

    crossmatch_data_dir = os.path.join(os.path.dirname(__file__), "data")
    crossmatch_files = {
        "news_sample": os.path.join(crossmatch_data_dir, "crossmatch_news_to_ztfnuc.json")
    }

    if base_name not in crossmatch_files:
        raise KeyError(f"No crossmatch file given for {base_name}")

    logger.debug(f"loading {crossmatch_files[base_name]}")
    with open(crossmatch_files[base_name], "r") as f:
        crossmatch = json.load(f)

    allwise_id = wise_data.parent_sample.df.iloc["Allwise_id", index]
    logger.debug(f"AllWISE ID: {allwise_id}")

    if allwise_id not in crossmatch:
        logger.debug("No match")
        return None

    return crossmatch[allwise_id]


def get_ztf_observability(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index | None = None,
        status: Status | None = None
) -> pd.DataFrame:

    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)

    if index is None:
        if status is None:
            raise ValueError("Either of index or status must be given!")
        else:
            indices = database_connector.get_ids(status)
    else:
        indices = list(np.atleast_1d(index))

    # -------------     spacial observability: Dec above -30 degrees      -------------- #
    #              as motivated by https://www.ipac.caltech.edu/news/dr12
    dec = wise_data.parent_sample.df.loc[np.array(indices).astype(int)][wise_data.parent_sample.default_keymap["dec"]]
    sky_vis = dec >= -29
    sky_vis.index = sky_vis.index.astype(str)
    logger.info(f"{sum(sky_vis)} in ZTF FoV")

    # ------------- temporal observability: peak date after 17 March 2018 -------------- #
    #            as also motivated by https://www.ipac.caltech.edu/news/dr12
    peak_times = database_connector.get_peak_time_jd(index=indices)
    ztf_start_jd = ztf_start.jd
    temporal_vis_per_band = peak_times >= ztf_start_jd
    temporal_vis = temporal_vis_per_band.any(axis=1)
    logger.info(f"{sum(temporal_vis)} in ZTF era")

    # -------------      observability: in ZTF sky during ZTF period      -------------- #
    vis = sky_vis & temporal_vis
    logger.info(f"{sum(vis)} in ZTF FoV and era")

    return vis


def ztf_detected(
        base_name: str,
        database_name: str,
        wise_data: WISEDataBase,
        index: Index,
        snr_threshold: float = 5,
        n_detections: int = 1,
        time_interval_relative_to_peak: tuple[float, float] = (-365, 2*365),
) -> pd.DataFrame:

    vis = get_ztf_observability(base_name, database_name, wise_data, index)
    vis_indices = vis.index[vis]

    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    peak_times = database_connector.get_peak_time_jd(index=vis_indices).median(axis=1)

    detected = dict()

    for i in tqdm(vis_indices, desc="going through forced photometry"):

        i_peak_time = Time(peak_times.loc[i], format="jd")

        lc = get_ztf_lightcurve(
            base_name=base_name,
            index=i,
            snr_threshold=snr_threshold,
        )

        if lc is None:
            continue

        lc_times = lc["obsmjd"]
        time_mask = (
                (lc_times >= i_peak_time.mjd + time_interval_relative_to_peak[0]) &
                (lc_times <= i_peak_time.mjd + time_interval_relative_to_peak[1])
        )

        idetected = dict()
        for f in ["g", "r", "i"]:
            if lc is not None:
                bandmask = lc["filter"] == f"ZTF_{f}"

                idetected[f] = np.sum(lc["above_snr"][bandmask] & time_mask) >= n_detections
            else:
                pass
                # idetected[f] = False

        detected[i] = idetected

    return pd.DataFrame.from_dict(detected, orient="index")



import argparse
import json
import logging
import subprocess
from timewise.wise_data_base import WISEDataBase

from timewise_sup.ampel_conf import ampel_conf_filename
from timewise_sup.analyse_lightcurves.create_job_file_yaml import make_ampel_job_file

logger = logging.getLogger("timewise_sup.analyze_lightcurves.bayesian_blocks")


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("base_name", type=str)
    parser.add_argument("chunk", type=int)
    parser.add_argument("--t2units", type=str, nargs="+")
    parser.add_argument("--precut_filter", type=bool, default=False, const=True, nargs="?")
    parser.add_argument("--logging-level", "-l", type=str, default="WARNING")
    return parser


def run_bayesian_blocks():
    cfg = vars(get_parser().parse_args())

    logging_level = cfg.pop("logging_level")

    # logging level might be something like `10` in which case we have to transform to an integer
    try:
        logging_level = int(logging_level)
    except ValueError:
        pass

    logging.getLogger("timewise_sup").setLevel(logging_level)
    logger.setLevel(logging_level)
    logger.debug(json.dumps(cfg, indent=4))
    bayesian_blocks(**cfg)


def keys_in_list(keys, arr):
    return any([any([k in element for element in arr]) for k in keys])


def bayesian_blocks(
    base_name: str,
    database_name: str,
    wise_data: WISEDataBase,
    chunk: int | None,
    t2units: list[str],
    load_from_bigdata_dir: bool = True,
    service: str = "tap",
    precut_filter: bool | None = False,
):

    fn = make_ampel_job_file(
        base_name=base_name,
        wise_data=wise_data,
        database_name=database_name,
        chunk=chunk,
        t2units=t2units,
        precut_filter=precut_filter,
        concurrent=False,
        split_units=False,
        load_from_bigdata_dir=load_from_bigdata_dir,
        service=service
    )

    logger.info("running ampel")
    with subprocess.Popen(["ampel", "job", "-config", ampel_conf_filename(), "-schema", fn]) as p:
        out, err = p.communicate()
        if out is not None:
            logger.info(out.decode())
        if err is not None:
            logger.error(err.decode())


if __name__ == "__main__":
    run_bayesian_blocks()



import logging
import pandas as pd

from timewise_sup.mongo import DatabaseConnector


logger = logging.getLogger(__name__)


def get_catalog_match_mask(
        base_name: str,
        database_name: str,
        index: list[str | int] | str | int
) -> pd.DataFrame:
    logger.info("getting catalog match mask")
    matches = DatabaseConnector(base_name=base_name, database_name=database_name).get_catalog_matches(index)
    has_match_dict = {i: {c: match is not None for c, match in imatches.items()} for i, imatches in matches.items()}
    has_match_masks = pd.DataFrame.from_dict(has_match_dict, orient="index").fillna(False)
    return has_match_masks



import logging
import os
import pandas as pd
import matplotlib.pyplot as plt

from timewise_sup.samples.sjoerts_flares import get_test_flares_config
from timewise_sup.plots.plot_lightcurves import bandcolors
from timewise_sup.environment import load_environment
from timewise_sup.meta_analysis.luminosity import get_ir_luminosities
from timewise_sup.plots import plots_dir


logger = logging.getLogger(__name__)
tsup_data_dir = load_environment("TIMEWISE_SUP_DATA")


def make_crosscheck_lightcurves_plots(
        show: bool = False
):
    logger.info("making crosscheck lightcurve plots")
    sjoerts_colors = {"W1": "darkred", "W2": "indianred"}

    # load data for flares from https://arxiv.org/abs/2111.09391
    test_config = get_test_flares_config()
    z = test_config.wise_data.parent_sample.df["z"]

    # get redshift and ids for flares where available
    zdict = dict()
    tpeak_dict = dict()
    for i, iz in z.items():
        try:
            zdict[i] = float(iz)
            tpeak_dict[i] = float(test_config.wise_data.parent_sample.df.loc[i, "tpeak"])
        except ValueError:
            pass

    ids = tuple(zdict.keys())

    # calculate luminosities
    lums = get_ir_luminosities(
        test_config.wise_data.base_name,
        test_config.database_name,
        test_config.wise_data,
        ids,
    )

    # plot the lightcurves for the three flares with lightcurve data in https://arxiv.org/abs/2111.09391
    for name in ["AT2019dsg", "AT2019fdr", "AT2019aalc"]:
        logger.info(f"making plot for {name}")

        id = test_config.wise_data.parent_sample.df.index[test_config.wise_data.parent_sample.df.name == name][0]
        wise_lc = pd.DataFrame.from_dict(
            lums[id], orient="columns"
        )
        tpeak = tpeak_dict[id]

        fig, ax = plt.subplots()

        for b in ["W1", "W2"]:
            ax.errorbar(
                wise_lc.mean_mjd - tpeak,
                wise_lc[f"{b}_luminosity_erg_per_s"],
                yerr=wise_lc[f"{b}_luminosity_err_erg_per_s"],
                marker='s',
                ecolor="k",
                capsize=2,
                ls="",
                color=bandcolors[b],
                label=f"WISE {b} (AIR-FLARES)",
                zorder=5,
                barsabove=True
            )

            # load the reference data
            without_at = name.replace("AT", "")
            ref_fn = os.path.join(tsup_data_dir, f"neutrino_flares/{without_at}_{b}.csv")
            logger.debug(f"loading {ref_fn}")
            ref = pd.read_csv(
                ref_fn,
                sep=";",
                decimal=",",
                names=["time_from_peak", "luminosity"]
            )

            ax.scatter(
                ref.time_from_peak,
                ref.luminosity * 1e43,
                marker='x',
                s=30,
                color=sjoerts_colors[b],
                label=f"WISE {b} (Sjoert)",
                zorder=5
            )

        ax.set_xlabel(f"t - {tpeak} [d]")
        ax.set_ylabel(r"Luminosity [erg s$^{-1}$]")
        ax.set_title(name)
        ax.legend()
        ax.grid()

        fn = os.path.join(plots_dir("crosscheck_lightcurves", "crosscheck"), f"{name}_crosscheck.pdf")
        logger.info(f"saving under {fn}")
        fig.savefig(fn)

        if show:
            plt.show()

        plt.close()



import logging


main_logger = logging.getLogger('timewise_sup')
logger_format = logging.Formatter('%(levelname)s:%(name)s:%(funcName)s - %(asctime)s: %(message)s', "%H:%M:%S")
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(logger_format)
main_logger.addHandler(stream_handler)
main_logger.propagate = False  # do not propagate to root logger



import logging
import os
import matplotlib.pyplot as plt
from astropy.time import Time
import numpy as np
from timewise import WISEDataDESYCluster

from timewise_sup.samples.sjoerts_flares import get_test_flares_config, sjoerts_base_name, sjoerts_database_name
from timewise_sup.meta_analysis.baseline_subtraction import get_baseline_subtracted_lightcurves, get_lightcurves
from timewise_sup.mongo import DatabaseConnector, Index, Status
from timewise_sup.plots import plots_dir


logger = logging.getLogger(__name__)


def make_lightcurves_ensemble_plot(
        base_name: str,
        database_name: str,
        wise_data: WISEDataDESYCluster,
        status: Status
):
    logger.info(f"making ensemble plot for {base_name} in {database_name} (status {status})")

    # get difference photometry lightcurves
    diff_lcs = get_baseline_subtracted_lightcurves(base_name, database_name, wise_data, status)

    # get peak times
    database_connector = DatabaseConnector(base_name=base_name, database_name=database_name)
    sample_ids = database_connector.get_ids(status)
    peak_t = database_connector.get_peak_time_jd(sample_ids)
    # move to MJD
    for b in peak_t.columns:
        peak_t[f"{b}_mjd"] = Time(peak_t[b], format="jd").mjd

    # try to get flares from https://arxiv.org/abs/2111.09391
    sjoerts_sample = get_test_flares_config().wise_data
    m = np.array([n in ["AT2019dsg", "AT2019fdr"] for n in sjoerts_sample.parent_sample.df.name])
    sjoert_ids = list(sjoerts_sample.parent_sample.df.index[m])
    sjoert_lcs = get_lightcurves(sjoerts_base_name, sjoerts_database_name, sjoerts_sample, sjoert_ids)
    sjoerts_connector = DatabaseConnector(base_name=sjoerts_base_name, database_name=sjoerts_database_name)
    sjoerts_peaks = sjoerts_connector.get_peak_time_jd(sjoert_ids)

    # make plot

    e = list()
    fig, axs = plt.subplots(
        nrows=2,
        sharex="all",
        figsize=[5, 7],
        sharey="all",
        gridspec_kw={"hspace": 0}
    )

    for f, ax in zip(["W1", "W2"], axs):
        logger.debug(f"plotting {f}")

        for i, (k, v) in enumerate(diff_lcs.items()):

            try:
                ipeak = peak_t.loc[k, f"max_mag_jd_{f}_mjd"]
                mjds = np.array(list(v["mean_mjd"].values()))
                diff_f = np.array(list(v[f"{f}_diff_mean_flux_density"].values()))
                label = f"AIR-FLARES (status {status})" if i == 0 else ""
                ax.plot(mjds - ipeak, diff_f, color="grey", marker="", alpha=0.1, label=label)

            except KeyError:
                e.append(k)

        for k, v in sjoert_lcs.items():
            ipeak = Time(sjoerts_peaks.loc[k, f"max_mag_jd_{f}"], format='jd').mjd
            mjds = np.array(list(v["mean_mjd"].values()))
            diff_f = np.array(list(v[f"{f}_diff_mean_flux_density"].values()))
            diff_f_e = np.array(list(v[f"{f}_diff_flux_density_rms"].values()))
            name = sjoerts_sample.parent_sample.df.loc[int(k), "name"]
            logger.debug(name)

            ax.errorbar(
                mjds - ipeak,
                diff_f,
                yerr=diff_f_e,
                marker="s",
                ms=3,
                capsize=2,
                label=name,
                ls=":"
            )

        ax.grid(":")
        ax.set_ylabel(f)

    if len(e) > 0:
        logger.warning(f"l{len(e)} KeyErrors! {e}")

    axs[0].legend()
    axs[0].set_ylim(1e-2, 80)
    axs[0].set_xlim(-1000, 1000)
    axs[0].set_yscale("log")
    axs[-1].set_xlabel("days from IR peak")
    fig.text(0.0, 0.5, r"$\nu$ F$_\nu$ [erg s$^{-1}$ cm$^{-2}$]", va='center', rotation='vertical')

    fn = os.path.join(plots_dir("ensemble_lightcurves", base_name), f"ensemble_lightcurves_status{status}.pdf")
    logger.info(f"saving under {fn}")
    fig.tight_layout()
    fig.savefig(fn)
    plt.close()
